﻿// Asteroids-like game - Jim Bumgardner
//

doJitter = true;			// set this to false if you don't like the jitter effect

MW = 8;						// Scaling factor for models (which were originally drawn on graph paper)
SW = Stage.width;			// Stage coords
SH = Stage.height;
kLineHeight = 16; 			// Height of lines on paper
kDegToRad = Math.PI/180;	// Useful constant for drawing circles & such
kDamp = 0.99;				// Damping in ship acceleration

initRocks = 3;				// Number of asteroids
kMaxSprites = 80;			// Maximum number of asteroids & bad guys
kMinSpriteLayer = 10;		// Base layer number for asteroids & bad guys
spriteIdx = 0;
spriteCount = 0;

kMaxBullets = 100;
kMinBulletLayer = 100;
nbrBullets = 0;
kLastBullet = 0;

nbrShards = 0;				// Shards are the lines you see when an object explodes
kMaxShards = 100;
kMinShardLayer = 200;

kMinTextLayer = 300;		// Texties are text sprites for text animations
nbrTexties = 0;
kMaxTexties = 100;

sprites = [];

score = 0;


// The models
//
// Rocket fuselage
fuseModel = [{mx:0, my:-5.5, x:0, y:-4, pen:.5, clr:0x000033, alpha:50},
			 {cx:-1,cy:-3,x:-1,y:-1, pen:2, clr:0x000033, alpha:50},
			 {x:-.75,y:4},
			 {cx:0, cy:4.5, x:.75,y:4},  // from -.75,4
			 {x:1,y:-1},
			 {cx:1,cy:-3,x:0,y:-4}];

// Rocket fins
finModel = [ {mx:-1,my:-1,cx:-3,cy:4,x:-2,y:6,bf:0x0000FF,bfa:80},
			 {cx:-1,cy:4.5,x:-.75,y:4}, 
			 {x:-1,y:-1},
			 {mx:.75,my:4,cx:1,cy:4.5,x:2,y:6},
			 {cx:3,cy:4,x:1,y:-1},
			 {x:.75,y:4,ef:1}
			];

// Thrust Shapes
thrustModels = [
				[{mx:-1,my:4.5,x:-.5,y:5.5, bf:0xE0A7A8, bfa:50},
					{x:-.1,y:4.5},{x:0,y:6},
					{x:1,y:4.7},{x:-1,y:4.5,ef:1}],
        [{mx:-1,my:4.5,x:-.5,y:6.2, bf:0xE0A7D8, bfa:50},
					{x:-.2,y:4.7},{x:.3,y:6.5},
					{x:1,y:4.5},{x:-1,y:4.5,ef:1}],
			   [{mx:-1,my:4.5,x:-.5,y:7.2, bf:0xE00000, bfa:50},
					{x:-.2,y:4.7},{x:.3,y:7.5},
					{x:1,y:4.5},{x:-1,y:4.5,ef:1}]];

// Large Asteroid
rockModel = [{mx:1, my:1, x:0,y:2, bf:0, bfa:50},
			 {x:-1,y:2}, {x:-2,y:1}, {x:-3,y:0},	 {x:-3,y:-1},
			 {x:-2,y:-2}, {x:-1,y:-2}, {x:0,y:-3},	 {x:2,y:-3},
			 {x:2,y:-2}, {x:2,y:0},	 {x:1,y:1,ef:1}];

// Small Asteroids (made from large asteroid)
subrockModels = [[{mx:2,my:1,x:1,y:1,bf:0, bfa:50},
				  {x:1,y:2},{x:0,y:2},{x:-1,y:1},{x:-1,y:0},{x:0,y:-1},{x:1,y:-1},{x:1,y:0},{x:2,y:1,ef:1}],
				  [{mx:0,my:2,x:-1,y:1,bf:0, bfa:50},
				   {x:-1,y:0},{x:0,y:-1},{x:2,y:-1},{x:2,y:0},{x:1,y:0},{x:1,y:1},{x:0,y:1},{x:0,y:2,ef:1}],
				  [{mx:-1,my:0,x:-1,y:-1,bf:0, bfa:50},
				   {x:0,y:-1},{x:0,y:-2},{x:1,y:-2},{x:1,y:0},{x:0,y:1},{x:-1,y:0,ef:1}],
				  [{mx:0,my:-1,x:1,y:0,bf:0, bfa:50},
				   {x:0,y:1},{x:-1,y:1},{x:-2,y:0},{x:-1,y:0},{x:-1,y:-1},{x:0,y:-1,ef:1}]];

// UFO
ufoModels = [[{mx:-4,my:0,x:4,y:0,bf:0xFF00FF, bfa:80}, // upper disc
			  {cx:3.5,cy:-1,x:1,y:-1},
			  {x:-1,y:-1},
			  {cx:-3.5,cy:-1,x:-4,y:0, ef:1}],
             [{mx:4,my:0,x:-4,y:0,bf:0x770077, bfa:80}, // lower disc
			  {cx:-1,cy:1,x:0,y:1},
			  {cx:1, cy:1,x:4,y:0,ef:1}],
			 [{mx:-1,my:-1,cx:0,cy:-.33,x:1,y:-1,bf:0xFFFF00, bfa:100},
			  {cx:0,cy:-2,x:-1,y:-1,ef:1}]];

bulletModel = [{mx:-8,my:0,cx:0,cy:0,x:0,y:-8,bf:0xcccc00,bfa:80},
								{cx:0,cy:0,x:8,y:0},
								{cx:0,cy:0,x:0,y:8},
								{cx:0,cy:0,x:-8,y:0, ef:1}];

// Routine to scale model to arbitrary size
scaleModel = function(m,s)
{
	for (var i = 0; i < m.length; ++i)
	{
		var pt = m[i];
		if (pt.mx != undefined)
		{
			pt.mx *= s;
			pt.my *= s;
		}
		if (pt.cx != undefined)
		{
			pt.cx *= s;
			pt.cy *= s;
		}
		pt.x *= s;
		pt.y *= s;
	}
}

scaleModel(fuseModel, MW*.8); // Using a slightly smaller rocket than original drawing
scaleModel(finModel, MW*.8);
scaleModel(thrustModels[0], MW*.8);
scaleModel(thrustModels[1], MW*.8);
scaleModel(thrustModels[2], MW*.8);
scaleModel(rockModel, MW);
scaleModel(subRockModels[0], MW);   
scaleModel(subRockModels[1], MW);   
scaleModel(subRockModels[2], MW);   
scaleModel(subRockModels[3], MW);   
scaleModel(ufoModels[0], MW);
scaleModel(ufoModels[1], MW);
scaleModel(ufoModels[2], MW);

MovieClip.prototype.drawCircle = function(x,y,radius)
{
	var r = radius;
	var n= 8;
	
	var theta = 45*kDegToRad;
	var cr = radius/Math.cos(theta/2);
	var angle = bA*kDegToRad;
	var cangle = angle-theta/2;
	
	this.moveTo(x+r, y);
	for (var i=0;i < n;i++) 
	{
		angle += theta;
		cangle += theta;
		var endX = r*Math.cos (angle);
		var endY = r*Math.sin (angle);
		var cX = cr*Math.cos (cangle);
		var cY = cr*Math.sin (cangle);
		this.curveTo(x+cX,y+cY, x+endX,y+endY);
	}
}

// One time routine to draw lined-paper background
MovieClip.prototype.drawPaper = function()
{
	this.clear();
	this.lineStyle(.5, 0xb9c4df, 100);
	for (var i =3 ; i < SH/kLineHeight; ++i) {
		this.moveTo(0,i*kLineHeight);
		this.lineTo(SW, i*kLineHeight);
	}
	this.lineStyle(.5, 0xE0D7D8, 100);
	this.moveTo(50, 0);
	this.lineTo(50, SH);

	this.lineStyle(.5, 0xCFD1D0, 50);
	this.drawCircle(25, 50, 8);
	this.drawCircle(25, 250, 8);
	this.drawCircle(25, 450, 8);

	this.holes.clear();
	this.holes.beginFill(0xFFFFFF, 100);
	this.holes.drawCircle(25, 50, 7);
	this.holes.drawCircle(25, 250, 7);
	this.holes.drawCircle(25, 450, 7);
	this.holes.endFill();
}

// This function creates a set of shard movie clips which correspond
// to the lines in a model - used for sprites, ship and ufo.
MovieClip.prototype.explode = function(duration)
{
	var m = this.model;
	// this._visible = false;
	var lx;
	var ly;
	var clr;
	for(i = 0; i < m.length; ++i) 
	{
		var pt = m[i];
		var mc = _root.createEmptyMovieClip("shard"+nbrShards, kMinShardLayer+nbrShards);
		nbrShards = (nbrShards+1)%kMaxShards;
		var opt = {x:0,y:0};
		this.localToGlobal(opt);
		mc._x = opt.x;
		mc._y = opt.y;
		mc._rotation = this._parent._rotation;
		if (mc._rotation == 0)
			mc._rotation = this._rotation;
		if (pt.clr != undefined)
			clr = pt.clr;
		else if (pt.bf != undefined)
			clr = pt.bf;
		if (pt.mx != undefined) {
			lx = pt.mx;
			ly = pt.my;
		}
		mc.clear();
		mc.lineStyle(2,clr, 100);
		mc.moveTo(lx,ly);
		if (pt.cx != undefined) {
			mc.curveTo(pt.cx,pt.cy,pt.x,pt.y);
		}
		else {
			mc.lineTo(pt.x,pt.y);
		}
		var dpt = {x:(lx + pt.x)/2, y:(ly + pt.y)/2};
		mc.localToGlobal(dpt);
		var dx = dpt.x - opt.x;
		var dy = dpt.y - opt.y;
		var dist = Math.sqrt(dx*dx+dy*dy);
		mc.vx = dx/dist + this.vx;
		mc.vy = dy/dist + this.vy;
		mc.n = 0;
		mc.startTime = getTimer();
		mc.explodeTime = duration;
		lx = pt.x;
		ly = pt.y;
		mc.spin = Math.random()*6-3;
		mc.onEnterFrame = function()
		{
			var r = (getTimer() - this.startTime)/this.explodeTime;
			if (r > 1) {
				this.onEnterFrame = undefined;
				this.removeMovieClip();
			}
			this._alpha = (1-r*r)*50;
			this._rotation += this.spin;
			this._x += this.vx;
			this._y += this.vy;
		}
	}
}

// Collision checking for ship - we check against the coordinates in the model
// This works better than using hitTest on ship_mc, which checks the bounding box only.
MovieClip.prototype.didHit = function(testObj, m)
{
	for (var i = 0; i < m.length; ++i) 
	{
		var pt = {x:m[i].x, y:m[i].y};
		testObj.localToGlobal(pt);
		if (this.hitTest(pt.x,pt.y,true))
			return true;
	}
	return false;
}

// Assign a sprite model to a movieclip & draw it
//
MovieClip.prototype.setModel = function(m)
{
	this.model = m;
	this.drawModel(m);
}

// Draw a model
//
drawModel = function(m)
{
	if (m == undefined)
		m = this.model;
	for (var i = 0; i < m.length; ++i) 
	{
		var pt = m[i];
		if (pt.bf != undefined)
			this.beginFill(pt.bf, pt.bfa);
		if (pt.pen != undefined)
			this.lineStyle(pt.pen,pt.clr,pt.alpha);
		if (pt.mx != undefined)
			this.moveTo(pt.mx,pt.my);
		if (pt.cx != undefined)
			this.curveTo(pt.cx,pt.cy,pt.x,pt.y);
		else if (pt.x != undefined)
			this.lineTo(pt.x,pt.y);
		if (pt.ef != undefined)
			this.endFill();
	}
}

drawModelJitter = function(m)
{
	if (m == undefined)
		m = this.model;
	for (var i = 0; i < m.length; ++i) 
	{
		var pt = m[i];
		if (pt.bf != undefined)
			this.beginFill(pt.bf, pt.bfa+Math.random()*20-10);
		if (pt.pen != undefined)
			this.lineStyle(pt.pen,pt.clr,pt.alpha+Math.random()*20-10);
		if (pt.mx != undefined)
			this.moveTo(pt.mx,pt.my);
		if (pt.cx != undefined)
			this.curveTo(pt.cx+Math.random()*2-1,pt.cy+Math.random()*2-1,pt.x+Math.random()*2-1,pt.y+Math.random()*2-1);
		else if (pt.x != undefined)
			this.lineTo(pt.x+Math.random()*2-1,pt.y+Math.random()*2-1);
		if (pt.ef != undefined)
			this.endFill();
	}
}

MovieClip.prototype.drawModel = doJitter? drawModelJitter : drawModel;

// Setup a text animation such as 'KABOOM'!
//
textAnimate = function(str,obj,ox,clr,dur)
{
	var n = ox;
	for (var i = 0; i < str.length; ++i) 
	{
		var ch = str.charAt(i);
		var mc = _root.attachMovie('letter_mc', 'char_'+nbrTexties, kMinTextLayer+nbrTexties, {_x:obj._x+n, _y:obj._y});
		nbrTexties = (nbrTexties+1)%kMaxTexties;
		mc.t_txt.autoSize = true;
		mc.t_txt.text = ch;
		mc.t_txt.textColor = clr;
		mc.vx = obj.vx;
		mc.vy = obj.vy;
		mc.dur = dur;
		mc._xscale = mc._yscale = 50;
		n += mc.t_txt.textWidth;
		mc.startTime = getTimer();
		mc.onEnterFrame = function()
		{
			var r = (getTimer() - this.startTime)/this.dur;
			if (r > 1)
				this.removeMovieClip();
			this._x += this.vx;
			this._y += this.vy;
			this._xscale = 50+100*r;
			this._yscale = 50+100*r;
			this._alpha = 100-(r*r)*100;
		};
	}
}

// Add to the score & display it
addScore = function(s)
{
	score += s;
	score_txt.text = "Score: " + score;
}

// Bullet movement
bulletMove = function()
{
	for (var i = 0; i < kMaxSprites; ++i) {
		if (sprites[i] != undefined && sprites[i]._visible && 
			sprites[i].hitTest(this._x, this._y,true))
		{
			this.removeMovieClip();
			sprites[i].gotHit();
			break;
		}
	}
	this._x += this.vx;
	this._y += this.vy;
	this._rotation -= 20;
	if (this.life++ > 100) {
		this.removeMovieClip();
	}
	if (doJitter) {
		this.clear();
		this.drawModel();
	}
}

// Launch a bullet when the user hits spacebar
//
launchBullet = function()
{
	if (getTimer() - kLastBullet < 300) // not too fast
		return;

	var mc = _root.createEmptyMovieClip("bullet"+nbrBullets, kMinBulletLayer+nbrBullets);
	mc.idx = nbrBullets;
	nbrBullets = (nbrBullets+1) % kMaxBullets;
	var vx = Math.sin(ship_mc._rotation*kDegToRad);
	var vy = -Math.cos(ship_mc._rotation*kDegToRad);
	mc._x = ship_mc._x+vx*6.5*MW;
	mc._y = ship_mc._y+vy*6.5*MW;
	mc.setModel(bulletModel);
	mc.vx = ship_mc.vx+vx*4;
	mc.vy = ship_mc.vy+vy*4;
	mc.life = 0;
	mc.onEnterFrame = bulletMove;
	kLastBullet = getTimer();
}

// Create a new sprite & keep track of it in an array
//
makeSprite = function()
{
	do {
		spriteIdx = (spriteIdx+1) % kMaxSprites;
	} while (sprites[spriteIdx] != undefined && spriteCount < kMaxSprites)
	var mc = _root.createEmptyMovieClip("rock_"+spriteIdx, kMinSpriteLayer+spriteIdx);
	sprites[spriteIdx] = mc;
	mc.idx = spriteIdx;
	spriteCount++;
	return mc;
}

// Kill a sprite
//
killSprite = function(spr)
{
	sprites[spr.idx] = undefined;
	spr.removeMovieClip();
	spriteCount--;
}


// Basic asteroid movement
//
rockMove = function()
{
	if (ship_mc._visible && (this.didHit(ship_mc, fuseModel) || this.didHit(ship_mc, finModel))) {
		ship_mc.gotHit();
	}
	this._rotation += this.spin;
	this._x += this.vx;
	this._y += this.vy;
	if (this._x < -this._width)
		this._x += SW+this._width*2;
	else if (this._x > SW+this._width)
		this._x -= SW+this._width*2;
	if (this._y < -this._height)
		this._y += SH+this._height*2;
	else if (this._y > SH+this._height)
		this._y -= SH+this._height*2;
	if (doJitter) {
		this.clear();
		this.drawModel();
	}
}

// What happens when we hit a small rock
//
smallRockHit = function()
{
	textAnimate("+"+this.points, this,-30,0xFF0000,2000);
	addScore(this.points);
	this.explode(2*1000);
	killSprite(this);
}

// What happens when we hit a big rock (it makes 4 small rocks)
//
rockHit = function()
{
	textAnimate("+"+this.points, this,-30,0xFF0000,2000);
	addScore(this.points);

	// break rock into 4 shards aligned to big rock
	for (var i = 0; i < 4; ++i) 
	{
		var mc = makeSprite();
		mc.setModel(subrockModels[i]);
		var dpt = {x:-mc.model[0].mx,y:-mc.model[0].my};
		this.localToGlobal(dpt);
		mc._x = dpt.x;
		mc._y = dpt.y;
		var dx = dpt.x - this._x;
		var dy = dpt.y - this._y;
		var dist = Math.sqrt(dx*dx+dy*dy);
		mc._rotation = this._rotation;
		mc.vx = this.vx+dx/dist;
		mc.vy = this.vy+dy/dist;
		mc.spin = (Math.random()*4-2);
		mc.onEnterFrame = rockMove;
		mc.gotHit = smallRockHit;
		mc.points = 200;
	}
	// remove original rock
	killSprite(this);
}

// Add a big rock to the stage
addRock = function()
{
	var mc = makeSprite();
	mc.setModel(rockModel);
	do {
		mc._x = random(SW);
		mc._y = random(SH);
		mc._rotation = random(360);
	} while (mc.hitTest(ship_mc));
	mc.vx = (Math.random()*4-2);
	mc.vy = (Math.random()*4-2);
	mc.spin = (Math.random()*4-2);
	mc.points = 150;
	mc.onEnterFrame = rockMove;
	mc.gotHit = rockHit;
}


// Add a few new rocks at the beginning of a level
//
NewRocks = function()
{
	for (var i = 0; i < initRocks; ++i)
	{
		addRock();
	}
	if (initRocks < 12)
		initRocks++;
}

// What happens when the UFO is hit
UFOHit = function()
{
	textAnimate("+"+this.points, this,-30,0xFF0000,2000);
	addScore(this.points);
	this.hidisc.explode(2*1000);
	this.lodisc.explode(2*1000);
	this.hood.explode(2*1000);
	killSprite(this);
}

// UFO Movement
UFOMove = function()
{
	if (ship_mc._visible && (this.didHit(ship_mc, fuseModel) || this.didHit(ship_mc, finModel))) {
		ship_mc.gotHit();
	}
	this._rotation = Math.sin(this.phase + getTimer()*.005)*6;
	this._x += this.vx;
	if (this._x > SW+this._width)
		killSprite(this);
	if (doJitter) {
		this.hidisc.clear(); this.hidisc.drawModel();
		this.lodisc.clear(); this.lodisc.drawModel();
		this.hood.clear();   this.hood.drawModel();
	}
}

// Make a new flying saucer
NewUFO = function()
{
	var mc = makeSprite();
	mc.createEmptyMovieClip("hidisc", 1);
	mc.createEmptyMovieClip("lodisc", 2);
	mc.createEmptyMovieClip("hood", 3);
	mc.hidisc.setModel(ufoModels[0]);
	mc.lodisc.setModel(ufoModels[1]);
	mc.hood.setModel(ufoModels[2]);
	mc._x = -mc._width;
	mc._y = mc._height + Math.random()*(SW-mc._height*2);
	mc.vx = 1 + random(4);
	mc.points = mc.vx*200;
	mc.phase = Math.random()*2*Math.PI;
	mc.gotHit = UFOHit;
	mc.onEnterFrame = ufoMove;
}

// Ship Movement and most game-play stuff happens here
ShipMove = function()
{
	if (spriteCount == 0) {
		// Level Up!!
		textAnimate("LEVEL UP!! +5000", this,-100,0x00FF00,4000);
		AddScore(5000);
		NewRocks();
	}

	if (random(500) == 0)
		NewUFO();

	this.clear(); // used to clear thrust drawings

	// Weapon firing
	if (Key.isDown(Key.SPACE))
	{
		launchBullet();
	}

	// Steering & Thrust
	if (Key.isDown(Key.LEFT))
		this._rotation -= 5;
	else if (Key.isDown(Key.RIGHT))
		this._rotation += 5;
	if (Key.isDown(Key.CONTROL) || Key.isDown(Key.UP))
	{
		this.vx += Math.sin(this._rotation*kDegToRad);
		this.vy -= Math.cos(this._rotation*kDegToRad);
		this.drawModel(thrustModels[random(3)]);
	}
	else if (Key.isDown(Key.DOWN)) 
	{
		this.vx *= .9;
		this.vy *= .9;
	}
	// Basic movement with acceleration and damping
	this.vx *= kDamp;
	this.vy *= kDamp;
	this._x += this.vx;
	this._y += this.vy;

	// Wrap around edges of stage
	if (this._x < -this._width)
		this._x += SW+this._width*2;
	else if (this._x > SW+this._width)
		this._x -= SW+this._width*2;
	if (this._y < -this._height)
		this._y += SH+this._height*2;
	else if (this._y > SH+this._height)
		this._y -= SH+this._height*2;
	if (doJitter) {
		this.fuselage.clear();
		this.fins.clear();
		this.fuselage.drawModel();
		this.fins.drawModel();
	}
	
}


// What happens when ship gets hit
shipHit = function()
{
	textAnimate("KABOOM!", this,-50,0xFF00FF,4000);
	this.fuselage.explode(3*1000);
	this.fins.explode(3*1000);
	this._visible = false;
	this.onEnterFrame = false; // Turn off game-play stuff
	// Wait 5 seconds and activate ship again
	this.si = function()
	{
		clearInterval(this.ni); 
		NewShip();
	};
	this.ni = setInterval(this, "si", 5*1000);
}

// Initialize new ship
function NewShip()
{
	ship_mc._x = SW/2;
	ship_mc._y = SH/2;
	ship_mc.vx = 0;
	ship_mc.vy = 0;
	ship_mc._visible = true;
	ship_mc._rotation = 0;
	ship_mc.onEnterFrame = shipMove;
	ship_mc.gotHit = shipHit;
}


// Key handling (all key states stored in array keys[])
//
_root.onKeyDown = function()
{
	var code = Key.getCode();
	switch (code) {
	case 69:	// shortcut E for testing ship explosions
		ship_mc.gotHit();
		break;
	case 68:	// shortcut D for testing UFO
		NewUFO();
		break;
	case 74:	// J turns jitter off
		doJitter = !doJitter;
		break;
	}
}


// One Time Initialziation
//
addScore(0); // initialize display score

_root.createEmptyMovieClip("paper_mc", 1);
paper_mc.createEmptyMovieClip("holes", 1);
paper_mc._x = 0;
paper_mc._y = 0;
paper_mc.drawPaper();

_root.createEmptyMovieClip("ship_mc", 2);
ship_mc.createEmptyMovieClip("fuselage", 1);
ship_mc.fuselage.setModel(fuseModel);
ship_mc.createEmptyMovieClip("fins", 2);
ship_mc.fins.setModel(finModel);
// ship_mc.drawModel();

NewShip();
NewRocks();
Key.addListener(_root);

